/*
 *  Kiss streaming server 
 *
 *  Copyright (C) 2003-2004 Thibault Lamy (xoubir@sourceforge.net)
 *                2003-2004 Jarl van Katwijk (bstard@sourceforge.net)
 *                2005      Erik Kunze (eku@sourceforge.net)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: clients.c,v 1.18 2005/08/18 06:38:45 eku Exp $
 */

#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <pthread.h>
#include <errno.h>
#include "dp500.h"
#include "logging.h"
#include "xmalloc.h"
#include "treat_event.h"
#include "shoutcast.h"
#include "clients.h"


static pthread_mutex_t mutex_client = PTHREAD_MUTEX_INITIALIZER;


static void *
serve_client(void *arg)
{
  t_dp500 *dp500 = gl_dp500;
  t_client *clt = arg;
  int fd = clt->fd;

  do_log2(4, fd, "entering thread to serve connection\n");

  while (clt->going)
    {
      struct timeval tv;
      fd_set rfds;

      tv.tv_sec = 0;
      tv.tv_usec = 20000;
      FD_ZERO(&rfds);
      FD_SET(fd, &rfds);

      if (select(fd + 1, &rfds, NULL, NULL, &tv) > 0)
	{
	  if (treat_event(clt, dp500) < 0)
	    {
	      del_client(clt, dp500);
	      break;		/* Client has closed connection */
	    }
	}
    }

  do_log2(4, fd, "leaving thread to serve connection\n");
  pthread_exit((void *)0);
}

static void
destroy_client(t_client * clt)
{
  int rc;

  do_log2(2, clt->fd, "closed connection\n");

  clt->going = 0;
  rc = pthread_join(clt->thread, NULL);
  if (rc && rc != EDEADLK)
    {
      do_log(1, "%s: pthread_join() failed: %s\n", __func__, strerror(rc));
    }
  shutdown(clt->fd, SHUT_RDWR);
  close(clt->fd);
  xfree(clt->buf);
  shoutcast_close(clt->stream);
  xfree(clt);
}

void
add_new_client(struct sockaddr_in *sa, int fd, t_dp500 * dp500)
{
  t_client *new;
  pthread_attr_t attr;
  int rc;

  /* Lock */
  rc = pthread_mutex_lock(&mutex_client);

  new = xmalloc(sizeof(t_client));
  memcpy(&new->sa, sa, sizeof(struct sockaddr_in));
  new->fd = fd;
  new->going = 1;
  new->action = 1;

  do_log2(4, fd, "creating thread to serve connection\n");
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  rc = pthread_create(&new->thread, &attr, serve_client, (void *)new);
  if (rc)
    {
      do_log(1, "%s: pthread_create() failed: %s\n", __func__, strerror(rc));
      destroy_client(new);
    }

  /* Add to list */
  new->next = dp500->clients;
  dp500->clients = new;

  /* Unlock */
  rc = pthread_mutex_unlock(&mutex_client);
}

void
del_client(t_client * clt, t_dp500 * dp500)
{
  t_client *tmp;
  t_client **pprec;
  int rc;

  /* Lock */
  rc = pthread_mutex_lock(&mutex_client);

  pprec = &dp500->clients;
  for (tmp = dp500->clients; tmp; tmp = tmp->next)
    {
      if (tmp == clt)
	{
	  *pprec = tmp->next;
	  destroy_client(tmp);
	  break;
	}
      pprec = &tmp->next;
    }

  /* Unlock */
  rc = pthread_mutex_unlock(&mutex_client);
}

t_client *
find_client(struct sockaddr_in *sa, t_dp500 * dp500)
{
  t_client *tmp = NULL;
  int rc;

  /* Lock */
  rc = pthread_mutex_lock(&mutex_client);

  for (tmp = dp500->clients; tmp; tmp = tmp->next)
    {
      if (memcmp(&sa->sin_addr, &tmp->sa.sin_addr, sizeof(struct in_addr)) ==
	  0)
	break;
    }

  /* Unlock */
  rc = pthread_mutex_unlock(&mutex_client);

  return tmp;
}

/* EOF */
