/*
 *  Kiss streaming server 
 *
 *  Copyright (C) 2003-2004 Thibault Lamy (xoubir@sourceforge.net)
 *                2003-2004 Jarl van Katwijk (bstard@sourceforge.net)
 *                2005      Erik Kunze (eku@sourceforge.net)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: cmd_get.c,v 1.33 2005/09/17 14:59:47 eku Exp $
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pthread.h>
#include <errno.h>
#include "dp500.h"
#include "logging.h"
#include "xmalloc.h"
#ifdef DVDSUPPORT
#include "dvd_tools.h"
#endif
#include "shoutcast.h"
#include "cmds.h"


#define MAX_BLOCK_SIZE 32767;


static void
cmd_get_stream(off_t offset, size_t len, t_client * clt)
{
  char *tmp;

  do_log2(4, clt->fd, "%s: offset=%lld len=%lld action=%d\n", __func__,
	  (unsigned long long)offset, (unsigned long long)len, clt->action);

  if (len == 0)
    len = MAX_BLOCK_SIZE;

  tmp = xmalloc(len);

  if (clt->action == 2)
    {
      /* DP500 is looking for ID3 => we return NULLs */
      do_log2(3, clt->fd, "%s: requested ID3 tag\n", __func__);
    }
  else
    {
      int new_len = shoutcast_read(tmp, (int)len, clt->stream);
      if (new_len != (int)len)
	{
	  do_log2(1, clt->fd, "%s: shoutcast_read(%i)=%i\n", __func__,
		  (int)len, new_len);
	  len = new_len;
	}
    }

  write(clt->fd, tmp, len);
  xfree(tmp);
}

/* Handle Kiss playback device GET requests. */
void
cmd_get(char *args, t_client * clt, t_dp500 * dp500)
{
  char *p1, *p2, *p3;
  const char *fullpath, *fext;
  const char *range;
  long o, l;
  off_t offset;
  size_t len;
  int filetype_audio = 0;
  int filetype_subp = 0;
  struct stat input_stat;
  int fd;
  char *buf = NULL;

  p1 = strchr(args, '|');
  p2 = strrchr(args, '\\');
  if (p1 == NULL)
    {
      do_log2(2, clt->fd, "%s: invalid request '%s'\n", __func__, args);
      return;
    }
  *p1 = '\0';
  if (p2 != NULL)
    *p2 = '/';
  p3 = strchr(p1 + 1, '|');
  if (p3 != NULL)
    *(p3 - 1) = '\0';
  else
    {
      p3 = strrchr(p1 + 1, ' ');
      if (p3 != NULL)
	*p3 = '\0';
    }

  fullpath = args;
  range = p1 + 2;

  do_log2(4, clt->fd, "%s: path '%s' range '%s'\n", __func__, fullpath,
	  range);

  if (sscanf(range, "%ld %ld", &o, &l) != 2)
    {
      do_log2(2, clt->fd, "%s: invalid range '%s'\n", __func__, range);
      goto out;
    }
  offset = o;
  len = l;

  /* Are we handling a streamed mp3? */
  if (strncmp(fullpath, "http://", 7) == 0)
    {
      cmd_get_stream(offset, len, clt);
      goto out;
    }

  /* Handle filename masquarading. Currently this is used for:
     - subpicture files that are not located in same directory as video
     - svcd images, which are extracted to a temporary file */
  do_log2(4, clt->fd, "%s: filename translation: real '%s' requested '%s'\n",
	  __func__, dp500->real_fullpath, dp500->requested_fullpath);

  if ((dp500->requested_fullpath != NULL) && (dp500->real_fullpath != NULL))
    {
      if (strcmp(fullpath, dp500->requested_fullpath) == 0)
	{
	  fullpath = dp500->real_fullpath;
	  do_log2(3, clt->fd,
		  "%s: using alternate file '%s' for requested '%s'\n",
		  __func__, dp500->real_fullpath, dp500->requested_fullpath);
	}
      else
	{
	  do_log2(3, clt->fd,
		  "%s: alternate file '%s' does not equal requested file '%s'\n",
		  __func__, dp500->requested_fullpath, fullpath);
	}
    }
  else
    {
      do_log2(3, clt->fd,
	      "%s: filenames not set, real '%s', requested '%s'\n", __func__,
	      dp500->real_fullpath, dp500->requested_fullpath);
    }

  /* Parse filename + path + extension */
  fext = strrchr(fullpath, '.');
  fext++;
  do_log2(4, clt->fd, "%s: sub extension '%s'\n", __func__, fext);

  if ((strcasecmp(fext, "mp3") == 0) || (strcasecmp(fext, "ogg") == 0) ||
      (strcasecmp(fext, "wma") == 0) || (strcasecmp(fext, "ac3") == 0))
    {
      do_log2(4, clt->fd, "%s: detected audio playback (.%s)\n", __func__,
	      fext);
      filetype_audio = 1;
    }
  else
    {
      if (strcasecmp(fext, "utf") == 0 || strcasecmp(fext, "sub") == 0 ||
	  strcasecmp(fext, "smi") == 0 || strcasecmp(fext, "rt ") == 0 ||
	  strcasecmp(fext, "txt") == 0 || strcasecmp(fext, "ssa") == 0 ||
	  strcasecmp(fext, "srt") == 0 || strcasecmp(fext, "aqt") == 0 ||
	  strcasecmp(fext, "jss") == 0)
	{
	  if (clt->path_subp_previous != NULL &&
	      strcmp(clt->path_subp_previous, fullpath) != 0)
	    {
	      xfree(clt->path_subp_previous);
	    }
	  if (clt->path_subp_previous == NULL && len == 0)
	    {
	      clt->path_subp_previous = xstrdup(fullpath);
	      clt->chunk_size = 35387;
	    }
	  else if (len == 0)
	    {
	      clt->chunk_size += 35387;
	    }
	  else
	    {
	      clt->chunk_size = len;
	    }

	  filetype_subp = 1;
	  len = clt->chunk_size;

	  do_log2(4, clt->fd,
		  "%s: detected subpicture playback (.%s) chunk=%i\n",
		  __func__, fext, clt->chunk_size);
	}
      else if (len == 0 && offset == 0)
	{
	  do_log2(4, clt->fd, "%s: detected picture playback (.%s)\n",
		  __func__, fext);
	}
      else
	{
	  do_log2(4, clt->fd, "%s: detected video playback (.%s)\n", __func__,
		  fext);
	}
    }

  /* Check if file exists */
  if (stat(fullpath, &input_stat) < 0)
    {
      int errno_sav = errno;

#ifdef DVDSUPPORT
      /* Are we reading from a dvd image? */
      if (strstr(fullpath, "DVDTITLE") != NULL)
	{
	  while (dvd_to_mpeg_read(offset, len, clt) == 1);
	}
      else
#endif
	{
	  do_log(1, "%s: stat(%s) failed: %s\n", __func__, fullpath,
		 strerror(errno_sav));
	}
      goto out;
    }

  /* len=0 is used to get all bytes for subpictures and .ogg files */
  if (len == 0)
    {
      len = input_stat.st_size;
    }

  do_log2(4, clt->fd, "%s: path '%s' len=%lld\n", __func__, fullpath,
	  (unsigned long long)len);

  fd = open(fullpath, O_RDONLY);
  if (fd < 0)
    {
      do_log(1, "%s: open(%s) failed: %s\n", __func__, fullpath,
	     strerror(errno));
      goto out;
    }

  /* Handle the action type */
  switch (clt->action)
    {
    case 1:
      /* ALL   : Normal block read */
      do_log2(4, clt->fd, "%s: ALL   : normal block read\n", __func__);
      offset = lseek(fd, offset, SEEK_SET);
      break;

    case 2:
      /* MP3   : Weird mode to get ID3 tag
         VIDEO : Telling us we're talking about subpictures */
      if (filetype_audio == 0)
	{
	  /* SUBPICTURES need 'normal' blockread .. ?? */
	  do_log2(4, clt->fd, "%s: VIDEO : subpicture read\n", __func__);
	  offset = lseek(fd, offset, SEEK_SET);
	}
      else
	{
	  /* MP3 tag need -len */
	  do_log2(4, clt->fd, "%s: MP3   : ID3 tag read\n", __func__);
	  offset = lseek(fd, -(off_t) len, SEEK_END);
	}
      break;

    default:
      /* VIDEO : Subpicture sync after a Search in video gets action numbers
         other than 1 or 2. */
      do_log2(2, clt->fd, "%s: unknown ACTION '%i', handling like '1'\n",
	      __func__, clt->action);
      offset = lseek(fd, offset, SEEK_SET);
    }

  if (offset < 0)
    {
      do_log(1, "%s: lseek() failed: %s\n", __func__, strerror(errno));
    }

  buf = xmalloc(len);
  if (read(fd, buf, len) < 0)
    {
      do_log(1, "%s: read() failed: %s\n", __func__, strerror(errno));
    }

  /* Writeout requested data to the Kiss playback device */
  do_log2(4, clt->fd, "%s: write %lld bytes\n", __func__,
	  (unsigned long long)len);
  write(clt->fd, buf, len);

  xfree(buf);
  close(fd);

out:
  return;
}

/* EOF */
