/*
 *  Kiss streaming server 
 *
 *  Copyright (C) 2003-2004 Thibault Lamy (xoubir@sourceforge.net)
 *                2003-2004 Jarl van Katwijk (bstard@sourceforge.net)
 *                2005      Erik Kunze (eku@sourceforge.net)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: files_tools.c,v 1.43 2005/09/16 10:17:17 eku Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "dp500.h"
#include "xmalloc.h"
#include "logging.h"
#include "playlists.h"
#include "shoutcast.h"
#include "pthread.h"
#include "common_menu.h"
#include "send_tools.h"
#include "files_tools.h"
#include "errno.h"
#include "string_replace.h"
#ifdef DVDSUPPORT
#include "dvd_tools.h"
#endif


#undef MARK_PLS
#undef MARK_M3U
#define MARK_PLAYLISTS


typedef int *(*f_source_handler) (char *sourcename_real);

typedef struct s_source_handler
{
  char *pattern;
  f_source_handler handler;
} t_source_handler;

typedef struct s_source_tree
{
  void *next;
  void *previous;

  const char *sourcename_kiss;
  const char *sourcename_path_kiss;
  const char *sourcename_real;
  const char *sourcename_path_real;

  t_source_handler *handler;
} t_source_tree;


static t_source_tree *source_tracker_tree = NULL;


/* All 'files' offered to the kiss should be added to dp500serv's internal
   source-tracker
   This function will 
     1. Prepare requested 'kiss' name for valid formatting
     2. Generate a kiss-name if sourcename_kiss = NULL
     3. Store the link between the real- and kiss-name in memory
     4. Add an handler to that real-kiss record for input handling (file, http,
        shoutcast, vcd extraction, etc)
*/
const char *
source_tracker_add(const char *sourcename_kiss,
		   const char *sourcename_path_real,
		   const char *sourcename_real)
{
  t_source_tree *stt_new;
  char temp[PATH_MAX];

#if 0
  do_log(4, "%s: kiss '%s', path '%s' real '%s'\n", __func__,
	 sourcename_kiss, sourcename_path_real, sourcename_real);
#endif

  /* Check if any realname is given */
  if (sourcename_real == NULL || strlen(sourcename_real) == 0)
    {
      return NULL;
    }

  /* Allocate new node in list */
  stt_new = xmalloc(sizeof(t_source_tree));
  if (sourcename_kiss == NULL || strlen(sourcename_kiss) == 0)
    {

      snprintf(temp, sizeof(temp), "%s\\%s", sourcename_path_real,
	       sourcename_real);
    }
  else
    {
      snprintf(temp, sizeof(temp), "%s\\%s", sourcename_path_real,
	       sourcename_kiss);
    }
  stt_new->sourcename_kiss = xstrdup(temp);
  snprintf(temp, sizeof(temp), "%s/%s", sourcename_path_real,
	   sourcename_real);
  stt_new->sourcename_real = xstrdup(temp);
  do_log(4, "%s: kiss '%s', real '%s'\n", __func__, stt_new->sourcename_kiss,
	 stt_new->sourcename_real);

  /* Now check for which handler to add, based on the real name 
     TODO replace this testing static code with filters the plugins will
     supply */
  stt_new->handler = NULL;

  /* TODO implement files handler, and link to that one to test */

  /* Check source for type
     1. Check extension (mpg like extension handling, image -> mpg)
     2. Check URL
     http://
     ftp://
     kiss://  (handler for this emulated kiss500 behavoir
     shoutcast://
     3. Check (S)VCD image
     4. Check for special (reboot, shutdown, etc)
     TODO
   */

  /* Set handler based on source type
     TODO */

  /* Add new record to the tracker list */
  if (source_tracker_tree == NULL)
    {
      /* Insert at the beginning of the list */
      source_tracker_tree = stt_new;
      stt_new->previous = NULL;
    }
  else
    {
      /* Insert at the end of the list */
      t_source_tree *stt_temp;

      for (stt_temp = source_tracker_tree; stt_temp != NULL;
	   stt_temp = stt_temp->next)
	{
	  if (stt_temp->next == NULL)
	    {
	      stt_temp->next = stt_new;
	      stt_new->previous = stt_temp;
	      break;
	    }
	}
    }

  return sourcename_kiss;
}

static void
source_tracker_clean(void)
{
  t_source_tree *stt_temp, *stt_temp_next;

  do_log(4, "%s: running\n", __func__);

  for (stt_temp = source_tracker_tree; stt_temp != NULL;
       stt_temp = stt_temp_next)
    {
      stt_temp_next = stt_temp->next;
      xfree(stt_temp->sourcename_kiss);
      xfree(stt_temp->sourcename_path_kiss);
      xfree(stt_temp->sourcename_real);
      xfree(stt_temp->sourcename_path_real);
      xfree(stt_temp);
    }
}

/* Returns realname (file/url/vcd image/etc) based on the Kiss name */
const char *
source_tracker_realname(const char *sourcename_kiss)
{
  t_source_tree *stt_temp;

  for (stt_temp = source_tracker_tree; stt_temp != NULL;
       stt_temp = stt_temp->next)
    {
      if (stt_temp->sourcename_kiss != NULL &&
	  strcmp(stt_temp->sourcename_kiss, sourcename_kiss) == 0)
	{
	  return stt_temp->sourcename_real;
	}
    }
  return NULL;
}

/* Returns kissname based of the realname */
const char *
source_tracker_kissname(const char *sourcename_real)
{
  t_source_tree *stt_temp;

  for (stt_temp = source_tracker_tree; stt_temp != NULL;
       stt_temp = stt_temp->next)
    {
      if (stt_temp->sourcename_real != NULL &&
	  strcmp(stt_temp->sourcename_real, sourcename_real) == 0)
	{
	  return stt_temp->sourcename_kiss;
	}
    }
  return NULL;
}

#if 0
/* Returns pointer to handler based on the Kiss name */
t_source_handler *
source_tracker_handler_by_kissname(const char *sourcename_kiss)
{
  t_source_tree *stt_temp;

  for (stt_temp = source_tracker_tree; stt_temp != NULL;
       stt_temp = stt_temp->next)
    {
      if (stt_temp->sourcename_kiss != NULL &&
	  strcmp(stt_temp->sourcename_kiss, sourcename_kiss) == 0)
	{
	  return stt_temp->handler;
	}
    }
  return NULL;
}

/* Setup the handler modules environment, load them, and query for handled
   pattern */
int
initialize_handlers(void)
{
  return 0;
}
#endif

static int
mycompar(const void *_a, const void *_b)
{
  struct dirent *a = *(struct dirent **)_a;
  struct dirent *b = *(struct dirent **)_b;

  if ((a->d_type == b->d_type) ||
      ((a->d_type == DT_LNK || a->d_type == DT_REG) &&
       (b->d_type == DT_LNK || b->d_type == DT_REG)))
    return (strcasecmp(a->d_name, b->d_name));
  return (a->d_type - b->d_type);
}

void
build_file_list_audio(const char *path, t_client * clt)
{
  char fullpath[PATH_MAX];
  char tmp[PATH_MAX];
  int is_playlist = 0;
  int numlist, n;
  struct dirent *de;
  struct dirent **namelist;
  struct stat st;

  if (strcmp(path, " ") != 0)
    {
      char *playlist_path, *available_path;

      snprintf(fullpath, sizeof(fullpath), "%s/%s", gl_dp500->mp3_path, path);

      /* If playlists is selected show the options. */
      if (strcmp(path, "Playlists_tmp") == 0)
	{
	  send_dir("-------------------  NEW PLAYLIST  ------------------",
		   "New_Playlist", clt);
	  send_dir("--------------  AVAILABLE PLAYLISTS  --------------",
		   "Show_Playlists", clt);
	  playlist_new_path();
	  goto out;
	}

      /* If new playlist is selected change the path and list all available
         entries from. */
      playlist_path = strstr(path, "Playlists_tmp/New_Playlist");
      if (playlist_path != NULL)
	{
	  char *p;
	  p = string_replace(path, "Playlists_tmp/New_Playlist", "");
	  snprintf(fullpath, sizeof(fullpath), "%s/%s", gl_dp500->mp3_path,
		   p);
	  xfree(p);
	  is_playlist = 1;
	}

      available_path = strstr(path, "Playlists_tmp/Show_Playlists");
      if (available_path != NULL)
	{

	  if (strcmp(path, "Playlists_tmp/Show_Playlists") == 0)
	    {
	      /* List all available playlists from the Playlist-dir */
	      playlist_show_playlists(clt);
	    }
	  else
	    {
	      /* Open a selected playlist */
	      char *p;
	      p = string_replace(path, "Playlists_tmp/Show_Playlists/", "");
	      snprintf(fullpath, sizeof(fullpath), "%s/%s",
		       gl_dp500->playlist_path, p);
	      xfree(p);
	      do_log2(3, clt->fd, "%s: %s\n", __func__, fullpath);
	      playlist_show_songs(fullpath, clt);
	    }
	  goto out;
	}
    }

  /* Show the Playlist-entry, but only in the root audio dir */
  else
    {
      strncpy(fullpath, gl_dp500->mp3_path, sizeof(fullpath) - 1);

      /* Send the virtual dir for common menu. */
      if (gl_dp500->show_server_menu == 1)
	send_common_menu_dir(clt);

      if (gl_dp500->enable_playlists == 1)
	send_dir("--------------------  PLAYLISTS  --------------------",
		 "Playlists_tmp", clt);
    }

  /* Open m3u files if they contain .m3u and are no dir. */
  if (strstr(fullpath, ".m3u") != NULL)
    {
      if (stat(fullpath, &st) < 0)
	{
	  do_log(1, "%s: stat(%s) failed: %s\n", __func__, fullpath,
		 strerror(errno));
	}
      else
	{
	  if (S_ISREG(st.st_mode))
	    {
	      playlist_show_songs(fullpath, clt);
	      goto out;
	    }
	}
    }

  do_log2(3, clt->fd, "%s: path '%s'\n", __func__, fullpath);
  numlist = scandir(fullpath, &namelist, NULL, mycompar);
  if (numlist < 0)
    {
      do_log(1, "%s: scandir(%s) failed: %s\n", __func__, fullpath,
	     strerror(errno));
      goto out;
    }
  for (n = 0; n < numlist; n++)
    {

      de = namelist[n];
      if (de->d_name[0] == '.')
	continue;

      snprintf(tmp, sizeof(tmp), "%s/%s", fullpath, de->d_name);
      if (stat(tmp, &st) < 0)
	{
	  do_log(1, "%s: stat(%s) failed: %s\n", __func__, tmp,
		 strerror(errno));
	  continue;
	}

      if (S_ISDIR(st.st_mode))
	{
#ifdef MARK_PLAYLIST
	  if (is_playlist == 1)
	    {
	      snprintf(tmp, sizeof(tmp), "-> %s", de->d_name);
	      send_dir(tmp, de->d_name, clt);
	    }
	  else
#endif
	    {
	      send_dir(de->d_name, de->d_name, clt);
	    }
	}
      else if (S_ISREG(st.st_mode))
	{
	  char *fext = strrchr(de->d_name, '.');
	  if (fext == NULL)
	    continue;
	  fext++;

	  if (strcasecmp(fext, "pls") == 0)
	    {
	      char *streamurl;
	      char *streamtitle;

	      streamurl = shoutcast_parse(tmp, &streamtitle);
	      if (streamurl)
		{
		  do_log2(3, clt->fd, "%s: stream found '%s'\n", __func__,
			  streamurl);
#ifdef MARK_PLS
		  snprintf(tmp, sizeof(tmp), "[STREAM] %s.mp3", streamtitle);
#else
#if 0
		  snprintf(tmp, sizeof(tmp), "%s.mp3", streamtitle);
#else
		  strncpy(tmp, de->d_name, sizeof(tmp) - 1);
		  tmp[sizeof(tmp) - 1] = '\0';
		  strcpy(strrchr(tmp, '.') + 1, "mp3");
#endif
#endif
		  streamurl = xrealloc(streamurl, strlen(streamurl) + 5);
		  strcat(streamurl, ".mp3");
		  send_item(tmp, NULL, streamurl, clt);
		  xfree(streamurl);
		  xfree(streamtitle);
		}
	    }
	  else if (strcasecmp(fext, "m3u") == 0)
	    {
	      char name_m3u[NAME_MAX + 7];
	      strncpy(tmp, de->d_name, sizeof(tmp) - 1);
	      tmp[strlen(tmp) - 4] = '\0';
	      snprintf(name_m3u, sizeof(name_m3u), "[M3U] %s", tmp);
#ifdef MARK_M3U
	      send_dir(name_m3u, de->d_name, clt);
#else
	      send_dir(tmp, de->d_name, clt);
#endif
	    }
	  else if ((strcasecmp(fext, "mp3") == 0) ||
		   (strcasecmp(fext, "ogg") == 0) ||
		   (strcasecmp(fext, "wma") == 0) ||
		   (strcasecmp(fext, "ac3") == 0))
	    {
#ifdef MARK_PLAYLIST
	      if (is_playlist == 1)
		{
		  char songname[NAME_MAX + 4];
		  snprintf(songname, sizeof(songname), "-> %s", de->d_name);
		  send_item(songname, fullpath, de->d_name, clt);
		}
	      else
#endif
		{
		  send_item(de->d_name, fullpath, de->d_name, clt);
		}
	    }
	}
    }
  for (n = 0; n < numlist; n++)
    xfree(namelist[n]);
  xfree(namelist);
out:
  write(clt->fd, "EOL\n", 4);
}

void
build_file_list_video(const char *path, t_client * clt)
{
  char fullpath[PATH_MAX];
  char tmp[PATH_MAX];
  int numlist, n;
  struct dirent *de;
  struct dirent **namelist;
#ifdef DVDSUPPORT
  int DVDparsed = FALSE;
#endif

  if (strcmp(path, " ") != 0)
    {
      snprintf(fullpath, sizeof(fullpath), "%s/%s", gl_dp500->video_path,
	       path);

      if (strcmp(path, RECENTLY_PLAYLIST_REALNAME) == 0)
	{
	  /* The Recently Played PLAYLIST */
	  if (playlist_export_recent(clt) != 1)
	    do_log(2, "%s: playlist_export_recent() failed\n", __func__);
	  goto out;
	}
    }
  else
    {
      strncpy(fullpath, gl_dp500->video_path, sizeof(fullpath) - 1);

      /* Send the virtual dir for common menu. */
      if (gl_dp500->show_server_menu == 1)
	send_common_menu_dir(clt);

      /* Add playlists */
      if (gl_dp500->enable_playlists == 1)
	send_dir(RECENTLY_PLAYLIST_KISSNAME, RECENTLY_PLAYLIST_REALNAME, clt);
    }

  do_log2(3, clt->fd, "%s: path '%s'\n", __func__, fullpath);
  numlist = scandir(fullpath, &namelist, NULL, mycompar);
  if (numlist < 0)
    {
      do_log(1, "%s: scandir(%s) failed: %s\n", __func__, fullpath,
	     strerror(errno));
      goto out;
    }
  for (n = 0; n < numlist; n++)
    {
      struct stat st;

      de = namelist[n];
      if (de->d_name[0] == '.')
	continue;

      snprintf(tmp, sizeof(tmp), "%s/%s", fullpath, de->d_name);
      if (stat(tmp, &st) < 0)
	{
	  do_log(1, "%s: stat(%s) failed: %s\n", __func__, tmp,
		 strerror(errno));
	  continue;
	}

      if (S_ISDIR(st.st_mode))
	{
	  send_dir(de->d_name, de->d_name, clt);
	}
      else if (S_ISREG(st.st_mode))
	{
	  char fixname[NAME_MAX];
	  char new_fext[5];
	  char *fext = strrchr(de->d_name, '.');
	  if (fext == NULL)
	    continue;
	  fext++;

	  /* Any supported extension? */
	  if (strcasecmp(fext, "avi") != 0 &&
	      strcasecmp(fext, "mpg") != 0 &&
	      strcasecmp(fext, "m2v") != 0 &&
	      strcasecmp(fext, "m2p") != 0 &&
	      strcasecmp(fext, "mpeg") != 0 &&
	      strcasecmp(fext, "vob") != 0 &&
#ifdef VCDSUPPORT
	      strcasecmp(fext, "bin") != 0 &&
	      strcasecmp(fext, "img") != 0 &&
#endif
#ifdef DVDSUPPORT
	      strcasecmp(fext, "ifo") != 0 &&
#endif
	      strcasecmp(fext, "wma") != 0)
	    continue;

	  /* Fix the extension to force DP500 play other formats like mpg */

	  /* Other mpg types */
	  if ((strcasecmp(fext, "m2v") == 0) ||
	      (strcasecmp(fext, "m2p") == 0))
	    {
	      strncpy(new_fext, "mpg", sizeof(new_fext));
	    }
#ifdef VCDSUPPORT
	  /* (S)VCD Images */
	  else if ((strcasecmp(fext, "bin") == 0) ||
		   (strcasecmp(fext, "img") == 0))
	    {
	      strncpy(new_fext, "mpg", sizeof(new_fext));
	      strcpy(fext, new_fext);
	    }
#endif
#ifdef DVDSUPPORT
	  /* Add titles parsed from an IFO file */
	  else if (strcasecmp(fext, "ifo") == 0)
	    {
	      if (DVDparsed == FALSE)
		{
		  detect_DVD_titles(fullpath, clt);
		  DVDparsed = TRUE;
		}
	      continue;
	    }
#endif
	  else
	    {
	      strncpy(new_fext, fext, sizeof(new_fext) - 1);
	    }

	  /* Fix the extension of the full path name to let playback
	     play other formats */
	  strncpy(fixname, de->d_name, sizeof(fixname) - 1);
	  fixname[sizeof(fixname) - 1] = '\0';
	  strcpy(strrchr(fixname, '.') + 1, new_fext);

	  /* Send line to playback device */
	  send_item(fixname, fullpath, de->d_name, clt);
	}
    }
  for (n = 0; n < numlist; n++)
    xfree(namelist[n]);
  xfree(namelist);
out:
  write(clt->fd, "EOL\n", 4);
}

void
build_file_list_images(const char *path, t_client * clt)
{
  char fullpath[PATH_MAX];
  char tmp[PATH_MAX];
  int numlist, n;
  struct dirent *de;
  struct dirent **namelist;

  if (strcmp(path, " ") != 0)
    snprintf(fullpath, sizeof(fullpath), "%s/%s", gl_dp500->images_path,
	     path);
  else
    {
      strncpy(fullpath, gl_dp500->images_path, sizeof(fullpath) - 1);

      /* Send the virtual dir for common menu. */
      if (gl_dp500->show_server_menu == 1)
	send_common_menu_dir(clt);
    }

  do_log2(3, clt->fd, "%s: path '%s'\n", __func__, fullpath);
  numlist = scandir(fullpath, &namelist, NULL, mycompar);
  if (numlist < 0)
    {
      do_log(1, "%s: scandir(%s) failed: %s\n", __func__, fullpath,
	     strerror(errno));
      goto out;
    }
  for (n = 0; n < numlist; n++)
    {
      struct stat st;

      de = namelist[n];
      if (de->d_name[0] == '.')
	continue;

      snprintf(tmp, sizeof(tmp), "%s/%s", fullpath, de->d_name);
      stat(tmp, &st);

      if (S_ISDIR(st.st_mode))
	{
	  send_dir(de->d_name, de->d_name, clt);
	}
      else if (S_ISREG(st.st_mode))
	{
	  char *fext = strrchr(de->d_name, '.');
	  if (fext == NULL)
	    continue;
	  fext++;

	  if (strcasecmp(fext, "jpg") == 0 || strcasecmp(fext, "jpeg") == 0)
	    {
	      send_item(de->d_name, fullpath, de->d_name, clt);
	    }
	}
    }
  for (n = 0; n < numlist; n++)
    xfree(namelist[n]);
  xfree(namelist);
out:
  write(clt->fd, "EOL\n", 4);
}

/* Scan path directory subtree for file.
   Returns newly allocated full_path on a hit, or NULL when nothing is found */
char *
scan_directory_tree(const char *path, const char *file)
{
  struct dirent *de;
  struct dirent **namelist;
  int numlist, n;
  const char *filename;
  char *result = NULL;

  /* Parse filename, strip path */
  filename = strrchr(file, '/');
  filename = filename == NULL ? file : filename + 1;

  do_log(4, "%s: path '%s' file '%s'\n", __func__, path, filename);

  /* Scan filesystem */
  numlist = scandir(path, &namelist, NULL, mycompar);
  if (numlist < 0)
    {
      do_log(1, "%s: scandir(%s) failed: %s\n", __func__, path,
	     strerror(errno));
      goto out;
    }

  for (n = 0; n < numlist; n++)
    {
      char tmp[PATH_MAX];
      struct stat st;

      de = namelist[n];
      if (de->d_name[0] == '.')
	continue;

      snprintf(tmp, sizeof(tmp), "%s/%s", path, de->d_name);
      if (stat(tmp, &st) < 0)
	{
	  do_log(1, "%s: stat(%s) failed: %s\n", __func__, tmp,
		 strerror(errno));
	  continue;
	}

      /* Sub dir scanning */
      if (S_ISDIR(st.st_mode))
	{
	  result = scan_directory_tree(tmp, filename);
	  if (result != NULL)
	    break;
	}
      /* Compare filename */
      if (strcmp(de->d_name, filename) == 0)
	{
	  result = xstrdup(tmp);
	  break;
	}
    }

  for (n = 0; n < numlist; n++)
    xfree(namelist[n]);
  xfree(namelist);

out:
  if (result)
    {
      do_log(4, "%s: '%s' found in '%s'\n", __func__, filename, result);
    }
  else
    {
      do_log(4, "%s: '%s' not found\n", __func__, filename);
    }

  return result;
}

/* EOF */
