/*
 *  Kiss streaming server 
 *
 *  Copyright (C) 2003-2004 Thibault Lamy (xoubir@sourceforge.net)
 *                2003-2004 Jarl van Katwijk (bstard@sourceforge.net)
 *                2001      Stephan Schmig
 *                2005      Erik Kunze (eku@sourceforge.net)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: playlists.c,v 1.16 2005/09/15 07:05:40 eku Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <errno.h>
#include "dp500.h"
#include "logging.h"
#include "xmalloc.h"
#include "send_tools.h"
#include "string_replace.h"
#include "playlists.h"


/* playlist_new_path selects a path to a new playlists. It begins with
   'Playlist 1.m3u'. If this file is already there it continues with
   'Playlist 2.m3u' and so on. */
int
playlist_new_path(void)
{
  char pl_path[PATH_MAX];
  int n;

  for (n = 1;; n++)
    {
      snprintf(pl_path, sizeof(pl_path), "%s/Playlist %i.m3u",
	       gl_dp500->playlist_path, n);

      if (access(pl_path, F_OK) != 0)
	{
	  xfree(gl_dp500->playlist_now);
	  gl_dp500->playlist_now = xstrdup(pl_path);
	  do_log(3, "New Playlistfile: %s\n", gl_dp500->playlist_now);
	  break;
	}
    }
  return 1;
}

/* playlist_show_songs lists songs of a m3u file and sends them to the player.
   All songs are being checked if accessable. If a mp3-file can not be opened
   it tries to convert the path of the song from windows path-names to unix
   path-names and tries again. */
int
playlist_show_songs(const char *fullpath, t_client * clt)
{
  int pl;
  struct stat pl_stat;
  int count, errno_save;
  char *pl_buffer = NULL;
  char *songpath = NULL, *songpath2 = NULL, *songpath3 = NULL;
  char *tokptr;

  do_log(2, "Playlist: opening '%s'\n", fullpath);

  pl = open(fullpath, O_RDONLY);
  if (pl < 0)
    {
      do_log(1, "%s: open(%s) failed: %s\n", __func__, fullpath,
	     strerror(errno));
      return 0;
    }

  fstat(pl, &pl_stat);
  pl_buffer = xmalloc(pl_stat.st_size + 1);
  count = read(pl, pl_buffer, (size_t) pl_stat.st_size);
  errno_save = errno;
  close(pl);
  if (count != pl_stat.st_size)
    {
      do_log(1, "%s: read() failed: %s\n", __func__, strerror(errno_save));
      return 0;
    }

  /* Convert windows m3u files, if needed */
  if (gl_dp500->convert_win_m3u_from &&
      strstr(pl_buffer, gl_dp500->convert_win_m3u_from) != NULL)
    {
      char *temp = string_replace(pl_buffer, gl_dp500->convert_win_m3u_from,
				  gl_dp500->convert_win_m3u_to);
      xfree(pl_buffer);
      pl_buffer = temp;
    }
  if (strstr(pl_buffer, "\\") != NULL)
    {
      char *temp = string_replace(pl_buffer, "\\", "/");
      xfree(pl_buffer);
      pl_buffer = temp;
    }
  if (strstr(pl_buffer, "\r") != NULL)
    {
      char *temp = string_replace(pl_buffer, "\r", "");
      xfree(pl_buffer);
      pl_buffer = temp;
    }

  songpath2 = xmalloc(PATH_MAX);
  songpath3 = xmalloc(PATH_MAX);
  for (songpath = strtok_r(pl_buffer, "\n", &tokptr); songpath != NULL;
       songpath = strtok_r(NULL, "\n", &tokptr))
    {
      char *songname;

      if (strstr(songpath, "#EXT") != NULL)
	{
	  continue;		/* ignore extended m3u tags */
	}

      /* Test if files are really there */
      if (access(songpath, R_OK) == 0)
	{
	  do_log(3, "Playlist: song send to player '%s'\n", songpath);
	  strncpy(songpath2, songpath, sizeof(songpath2) - 1);
	  songpath2[sizeof(songpath2) - 1] = '\0';
	}
      else
	{
	  /* If not found, try to make absolute paths */
	  snprintf(songpath2, PATH_MAX, "%s/%s", gl_dp500->mp3_path,
		   songpath);
	  if (access(songpath2, R_OK) != 0)
	    {
	      continue;
	    }
	  do_log(3, "Playlist: modified song send to player '%s'\n",
		 songpath2);
	}
      if (realpath(songpath2, songpath3) == NULL)
	{
	  do_log(1, "%s: realpath(%s) failed: %s\n", __func__, songpath2,
		 strerror(errno));
	}
      else
	{
	  songname = strrchr(songpath3, '/');
	  if (songname != NULL)
	    {
	      *songname++ = '\0';
	      send_item(songname, songpath3, songname, clt);
	    }
	}
    }
  xfree(songpath3);
  xfree(songpath2);
  xfree(pl_buffer);

  return 1;
}

/* playlist_show_playlists lists all .m3u files of the playlist folder in
   the KiSS "available playlists" folder. */
int
playlist_show_playlists(t_client * clt)
{
  int numlist, n;
  char *fext;
  struct dirent *de;
  struct dirent **namelist;

  numlist = scandir(gl_dp500->playlist_path, &namelist, NULL, alphasort);
  if (numlist < 0)
    {
      do_log(1, "%s: scandir(%s) failed: %s\n", __func__,
	     gl_dp500->playlist_path, strerror(errno));
      return 0;
    }

  for (n = 0; n < numlist; n++)
    {
      de = namelist[n];
      if (de->d_name[0] == '.')
	continue;

      fext = strrchr(de->d_name, '.');
      if (fext == NULL)
	continue;
      fext++;

      /* Send to player only if file is m3u */
      if (strcasecmp(fext, "m3u") == 0)
	{
	  char item_name[NAME_MAX];
	  strncpy(item_name, de->d_name, sizeof(item_name) - 1);
	  item_name[NAME_MAX - 1] = '\0';
	  /* Cut off the extension m3u for displaying */
	  item_name[strlen(item_name) - 4] = '\0';
	  send_dir(item_name, de->d_name, clt);
	}
    }
  for (n = 0; n < numlist; n++)
    xfree(namelist[n]);
  xfree(namelist);

  return 1;
}

/* playlist_insert inserts a song-path to the current active playlist. */
int
playlist_insert(const char *path)
{
  int pl, rc = 0;

  pl = open(gl_dp500->playlist_now, O_CREAT | O_WRONLY | O_APPEND);
  if (pl < 0)
    {
      do_log(1, "%s: open(%s) failed: %s\n", __func__, gl_dp500->playlist_now,
	     strerror(errno));
      goto out;
    }

  if (write(pl, path, strlen(path)) != (ssize_t) strlen(path) ||
      write(pl, "\n", 1) != 1)
    {
      do_log(1, "%s: write() failed: %s\n", __func__, strerror(errno));
      goto out;
    }
  rc = 1;

out:
  if (pl != -1)
    close(pl);

  return rc;
}

int
playlist_update_recent(const char *path)
{
#if 0
  char fext[4];
#endif
  char *file;
  char *pl_path = NULL;
  char *pl_line = NULL;
  size_t pl_line_len;
  int pl = -1;
  char *pl_buffer = NULL;
  char *pl_buffer_end = NULL;
  int found = 0;

  /* Build file entry */
  file = strrchr(path, '/');
  if (file == NULL)
    goto out;

#if 0
  strcpy(fext, file + strlen(file) - 4);	/* fetch extension */
  if ((strcmp(fext, ".mpg") == 0) || (strcpy(fext, "mpeg") == 0))
    {
      strcpy(file + strlen(file) - 3, "avi");	/* Set extension to avi */
    }
#endif
  pl_line = xmalloc(strlen(file) + 2 + strlen(path) + 5);
  sprintf(pl_line, "%s| %s|0|\n", file, path);
  pl_line_len = strlen(pl_line);

  /* Read existing recent.pl */
  pl_path = xmalloc(PATH_MAX);
  snprintf(pl_path, PATH_MAX, "%s/recent.pl", gl_dp500->playlist_path);
  pl = open(pl_path, O_RDONLY);
  if (pl != -1)
    {
      struct stat pl_stat;

      fstat(pl, &pl_stat);
      if (pl_stat.st_size > 0)
	{
	  char *p;

	  pl_buffer = xmalloc(pl_stat.st_size);
	  if (read(pl, pl_buffer, pl_stat.st_size) != pl_stat.st_size)
	    {
	      do_log(1, "%s: read() failed: %s\n", __func__, strerror(errno));
	      goto out;
	    }
	  pl_buffer_end = pl_buffer + pl_stat.st_size;

	  /* Search for current file already in list */
	  for (p = pl_buffer; p + pl_line_len < pl_buffer_end; p++)
	    {
	      if (strncmp(p, pl_line, pl_line_len) == 0)
		{
		  /* Hit */
		  do_log(4, "%s: found '%s' at pos '%i'\n", __func__, pl_line,
			 p - pl_buffer);
		  found = 1;
		  break;
		}
	    }
	}
    }
  else
    {
      do_log(4, "%s: open(%s) failed: %s\n", __func__, pl_path,
	     strerror(errno));
    }

  /* Write recent.pl if we have a new entry */
  if (found != 0)
    goto out;

  close(pl);
  pl = open(pl_path, O_CREAT | O_TRUNC | O_WRONLY);
  if (pl < 0)
    {
      do_log(1, "%s: open(%s) failed: %s\n", __func__, pl_path,
	     strerror(errno));
      goto out;
    }

  /* Write new one */
  if ((size_t) write(pl, pl_line, pl_line_len) != pl_line_len)
    {
      do_log(1, "%s: write() failed: %s\n", __func__, strerror(errno));
      goto out;
    }

  if (pl_buffer != NULL)
    {
      int lines = 0;
      size_t count;
      char *p;

      /* Shrink till RECENLTY_HISTORY amount of entries */
      for (p = pl_buffer; p < pl_buffer_end; p++)
	{
	  if (*p == '\n')
	    lines++;
	}
      do_log(4, "%s: lines in recently (1) '%i'\n", __func__, lines);

      if (lines > gl_dp500->playlist_recent_entries)
	{
	  for (--p; p >= pl_buffer; p--)
	    {
	      if (*p == '\n')
		{
		  lines--;
		  if (lines < gl_dp500->playlist_recent_entries)
		    {
		      p++;
		      break;
		    }
		}
	    }

	  do_log(4, "%s: lines in recently (2) '%i'\n", __func__, lines);
	}

      /* Write old entries */
      count = p - pl_buffer;
      if ((size_t) write(pl, pl_buffer, count) != count)
	{
	  do_log(1, "%s: write() failed: %s\n", __func__, strerror(errno));
	}
    }

out:
  if (pl != -1)
    close(pl);
  xfree(pl_buffer);
  xfree(pl_line);
  xfree(pl_path);
  return 1;
}

/* playlist_export_recent reads the recent playlist, and exports to playback
   device FILES are stored in [playlist_path]/recent.pl. */
int
playlist_export_recent(t_client * clt)
{
  char pl_path[PATH_MAX];
  int pl;

  /* The Recently Played PLAYLIST */
  snprintf(pl_path, sizeof(pl_path), "%s/recent.pl", gl_dp500->playlist_path);
  pl = open(pl_path, O_RDONLY);
  if (pl < 0)
    {
      write(clt->fd, "[EMPTY PLAYLIST].avi| [EMPTY PLAYLIST].avi|0|\n", 46);
      do_log(2, "PLAYLIST %s not found.\n", pl_path);
    }
  else
    {
      struct stat pl_stat;
      char *pl_buffer;
      size_t count;

      fstat(pl, &pl_stat);
      count = (size_t) pl_stat.st_size;
      pl_buffer = xmalloc(count);
      if (read(pl, pl_buffer, count) != (ssize_t) count)
	{
	  do_log(1, "%s: read() failed: %s\n", __func__, strerror(errno));
	}
      close(pl);
      write(clt->fd, pl_buffer, count);
      xfree(pl_buffer);
    }

  return 1;
}

/* EOF */
