/*
 *  Kiss streaming server 
 *
 *  Copyright (C) 2003-2004 Thibault Lamy (xoubir@sourceforge.net)
 *                2003-2004 Jarl van Katwijk (bstard@sourceforge.net)
 *                2005      Erik Kunze (eku@sourceforge.net)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: shoutcast.c,v 1.4 2005/09/17 14:58:54 eku Exp $
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <errno.h>
#include "dp500.h"
#include "logging.h"
#include "xmalloc.h"
#include "xconnect.h"
#include "shoutcast.h"


/* Shoutcast code taken from XMMS */

#define min(x,y) ((x)<(y)?(x):(y))
#define min3(x,y,z) (min(x,y)<(z)?min(x,y):(z))
#define min4(x,y,z,w) (min3(x,y,z)<(w)?min3(x,y,z):(w))

#define BASE64_LENGTH(len) (4 * (((len) + 2) / 3))


/* Encode the string S of length LENGTH to base64 format and place it
   to STORE.  STORE will be 0-terminated, and must point to a writable
   buffer of at least 1+BASE64_LENGTH(length) bytes.  */
static void
base64_encode(const char *s, char *store, size_t length)
{
  /* Conversion table.  */
  static char tbl[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
    'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
    'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
    'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
    'w', 'x', 'y', 'z', '0', '1', '2', '3',
    '4', '5', '6', '7', '8', '9', '+', '/'
  };
  size_t i;
  unsigned char *p = (unsigned char *)store;

  /* Transform the 3x8 bits to 4x6 bits, as required by base64.  */
  for (i = 0; i < length; i += 3)
    {
      *p++ = tbl[s[0] >> 2];
      *p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
      *p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
      *p++ = tbl[s[2] & 0x3f];
      s += 3;
    }

  /* Pad the result if necessary...  */
  if (i == length + 1)
    *(p - 1) = '=';
  else if (i == length + 2)
    *(p - 1) = *(p - 2) = '=';

  /* ...and zero-terminate it.  */
  *p = '\0';
}

/* Create the authentication header contents for the `Basic' scheme.
   This is done by encoding the string `USER:PASS' in base64 and
   prepending `HEADER: Basic ' to it.  */
static const char *
basic_authentication_encode(const char *user, const char *passwd,
			    const char *header)
{
  char *t1, *t2, *res;
  size_t len1 = strlen(user) + 1 + strlen(passwd);
  size_t len2 = BASE64_LENGTH(len1);

  t1 = xmalloc(len1 + 1);
  sprintf(t1, "%s:%s", user, passwd);
  t2 = xmalloc(len2 + 1);
  base64_encode(t1, t2, len1);
  res = xmalloc(strlen(header) + 10 + strlen(t2) + 1);
  printf(res, "%s: Basic %s\r\n", header, t2);
  xfree(t2);
  xfree(t1);

  return res;
}

static void
parse_url(const char *url, char **user, char **pass, char **host,
	  int *port, char **filename)
{
  char *h, *p, *pt, *f, *temp, *ptr;

  temp = xstrdup(url);
  ptr = temp;

  if (!strncasecmp("http://", ptr, 7))
    ptr += 7;
  h = strchr(ptr, '@');
  f = strchr(ptr, '/');
  if (h != NULL && (!f || h < f))
    {
      *h = '\0';
      p = strchr(ptr, ':');
      if (p != NULL && p < h)
	{
	  *p = '\0';
	  p++;
	  *pass = xstrdup(p);
	}
      else
	*pass = NULL;
      *user = xstrdup(ptr);
      h++;
      ptr = h;
    }
  else
    {
      *user = NULL;
      *pass = NULL;
      h = ptr;
    }
  pt = strchr(ptr, ':');
  if (pt != NULL && (f == NULL || pt < f))
    {
      *pt = '\0';
      *port = atoi(pt + 1);
    }
  else
    {
      if (f)
	*f = '\0';
      *port = 80;
    }
  *host = xstrdup(h);

  if (f)
    *filename = xstrdup(f + 1);
  else
    *filename = NULL;
  xfree(temp);
}

static int
http_used(t_mp3stream * pmst)
{
  if (pmst->wr_index >= pmst->rd_index)
    return pmst->wr_index - pmst->rd_index;
  return pmst->buffer_length - (pmst->rd_index - pmst->wr_index);
}

static int
http_free(t_mp3stream * pmst)
{
  if (pmst->rd_index > pmst->wr_index)
    return (pmst->rd_index - pmst->wr_index) - 1;
  return (pmst->buffer_length - (pmst->wr_index - pmst->rd_index)) - 1;
}

static void
http_wait_for_data(int length, t_mp3stream * pmst)
{
  while ((pmst->prebuffering || http_used(pmst) < length) && !pmst->eof
	 && pmst->going)
    usleep(10000);
}

static int
http_check_for_data(t_mp3stream * pmst)
{
  struct timeval tv;
  fd_set set;

  tv.tv_sec = 0;
  tv.tv_usec = 20000;
  FD_ZERO(&set);
  FD_SET(pmst->sock, &set);
  if (select(pmst->sock + 1, &set, NULL, NULL, &tv) > 0)
    return 1;
  return 0;
}

static int
http_read_line(char *buf, int size, t_mp3stream * pmst)
{
  int i = 0;

  while (pmst->going && i < size - 1)
    {
      if (http_check_for_data(pmst) == 0)
	continue;

      if (read(pmst->sock, buf + i, 1) <= 0)
	return -1;
      if (buf[i] == '\n')
	break;
      if (buf[i] != '\r')
	i++;
    }
  if (!pmst->going)
    return -1;
  buf[i] = '\0';
  return i;
}

static void *
http_buffer_loop(void *arg)
{
  char line[1024], *user, *pass, *host, *filename, *url, *temp, *file;
  const char *chost;
  int cnt, written, port, cport;
  int redirect;
  char service[6];

  t_dp500 *dp500 = gl_dp500;
  t_mp3stream *pmst = arg;
  url = xstrdup(pmst->url);

  do_log2(4, pmst->clt->fd, "entering thread to serve Shoutcast stream\n");

  do
    {
      const char *auth = NULL, *proxy_auth = NULL;

      redirect = 0;

      parse_url(url, &user, &pass, &host, &port, &filename);

      if ((!filename || *filename == '\0') && url[strlen(url) - 1] != '/')
	{
	  temp = xmalloc(strlen(url) + 2);
	  sprintf(temp, "%s/", url);
	  xfree(url);
	  url = temp;
	}

      chost = dp500->proxy_host ? dp500->proxy_host : host;
      cport = dp500->proxy_host ? dp500->proxy_port : port;

      snprintf(service, sizeof(service), "%d", cport);
      pmst->sock = xconnect(chost, service);
      if (pmst->sock == -1)
	{
	  pmst->eof = 1;
	  break;
	}

      if (user && pass)
	auth = basic_authentication_encode(user, pass, "Authorization");

      if (dp500->proxy_user)
	{
	  file = xstrdup(url);
	  if (dp500->proxy_user && dp500->proxy_password)
	    {
	      proxy_auth = basic_authentication_encode(dp500->proxy_user,
						       dp500->proxy_password,
						       "Proxy-Authorization");
	    }
	}
      else
	{
	  file = xmalloc(strlen(filename) + 2);
	  sprintf(file, "/%s", filename);
	}
      temp = xmalloc(1024);
      snprintf(temp, 1024,
	       "GET %s HTTP/1.0\r\n"
	       "Host: %s\r\n"
	       "User-Agent: %s/%s\r\n"
	       "%s%s\r\n"
	       "Icy-MetaData: 1\r\n",
	       file, host, PACKAGE, VERSION,
	       proxy_auth ? proxy_auth : "", auth ? auth : "");

      xfree(file);
      xfree(proxy_auth);
      xfree(auth);
      write(pmst->sock, temp, strlen(temp));
      xfree(temp);

      do_log2(4, pmst->clt->fd, "%s: connected, waiting for reply\n",
	      __func__);
      while (pmst->going && !pmst->eof)
	{
	  char *status;

	  if (http_check_for_data(pmst) == 0)
	    continue;

	  if (http_read_line(line, sizeof(line), pmst) <= 0)
	    {
	      pmst->eof = 1;
	      break;
	    }

	  status = strchr(line, ' ');
	  if (status)
	    {
	      if (status[1] == '2')
		break;
	      else if (status[1] == '3' && status[2] == '0' &&
		       status[3] == '2')
		{
		  while (pmst->going)
		    {
		      if (http_check_for_data(pmst) == 0)
			continue;

		      cnt = http_read_line(line, sizeof(line), pmst);
		      if (cnt == -1)
			{
			  pmst->eof = 1;
			  break;
			}
		      if (cnt == 0)
			break;
		      if (!strncmp(line, "Location:", 9))
			{
			  xfree(url);
			  url = xstrdup(line + 10);
			}
		    }
		  redirect = 1;
		  break;
		}
	      else
		{
		  do_log2(4, pmst->clt->fd, "%s: server response '%s'\n",
			  __func__, status);
		  break;
		}
	    }
	}

      while (pmst->going && !redirect)
	{
	  if (http_check_for_data(pmst) == 0)
	    continue;

	  cnt = http_read_line(line, sizeof(line), pmst);
	  if (cnt == -1)
	    {
	      pmst->eof = 1;
	      break;
	    }
	  if (cnt == 0)
	    break;
	  if (!strncmp(line, "icy-name:", 9))
	    {
	      xfree(pmst->icy_name);
	      pmst->icy_name = xstrdup(line + 9);
	    }
	  else if (!strncmp(line, "x-audiocast-name:", 17))
	    {
	      xfree(pmst->icy_name);
	      pmst->icy_name = xstrdup(line + 17);
	    }
	  if (!strncmp(line, "icy-metaint:", 12))
	    {
	      pmst->icy_metaint = atoi(line + 12);
	      do_log2(3, pmst->clt->fd, "%s: metaint=%i\n", __func__,
		      pmst->icy_metaint);
	    }
	  if (!strncmp(line, "x-audiocast-udpport:", 20))
	    {
	      do_log2(3, pmst->clt->fd,
		      "%s: server wants UDP messages on port %d\n", __func__,
		      atoi(line + 20));
	    }
	}

      if (redirect)
	{
	  close(pmst->sock);
	  xfree(user);
	  xfree(pass);
	  xfree(host);
	  xfree(filename);
	}
    }
  while (redirect);

  xfree(user);
  xfree(pass);
  xfree(host);
  xfree(filename);
  xfree(url);

  do_log2(4, pmst->clt->fd, "%s: now reading audio stream\n", __func__);

  while (pmst->going && !pmst->eof)
    {
      if (!http_used(pmst))
	pmst->prebuffering = 1;

      if (http_free(pmst) > 0 && !pmst->eof)
	{
	  if (http_check_for_data(pmst))
	    {
	      cnt = min(http_free(pmst),
			pmst->buffer_length - pmst->wr_index);
	      if (cnt > 1024)
		cnt = 1024;
	      written = read(pmst->sock, pmst->buffer + pmst->wr_index, cnt);
	      if (written <= 0)
		{
		  pmst->eof = 1;
		  if (pmst->prebuffering)
		    {
		      pmst->prebuffering = 0;
		    }
		}
	      else
		pmst->wr_index = (pmst->wr_index + written) %
		  pmst->buffer_length;
	    }

	  if (pmst->prebuffering)
	    {
	      if (http_used(pmst) > pmst->prebuffer_length)
		{
		  pmst->prebuffering = 0;
		}
#if 0
	      else
		{
		  do_log2(4, pmst->clt->fd, "%s: pre-buffering %dKB/%dKB\n",
			  __func__, http_used(pmst) / 1024,
			  pmst->prebuffer_length / 1024);
		}
#endif
	    }
	}
      else
	usleep(10000);
    }

  if (pmst->sock != -1)
    {
      close(pmst->sock);
      pmst->sock = -1;
    }
  do_log2(4, pmst->clt->fd, "%s: leaving thread to serve Shoutcast stream\n",
	  __func__);
  pthread_exit((void *)0);
}

char *
shoutcast_parse(const char *file, char **title)
{
  FILE *f;
  char *result = NULL;

  do_log(4, "%s: reading from '%s'\n", __func__, file);
  f = fopen(file, "r");
  if (f != NULL)
    {
      char buf[1024];

      while (fgets(buf, sizeof(buf), f))
	{
	  size_t len = strlen(buf);
	  buf[len - 1] = '\0';	/* remove newline */
	  if (len > 6 && strncmp(buf, "File1=", 6) == 0)
	    {
	      result = xstrdup(buf + 6);
	    }
	  if (len > 7 && strncmp(buf, "Title1=", 7) == 0)
	    {
	      if (title)
		*title = xstrdup(buf + 7);
	    }
	}
      fclose(f);
    }

  return result;
}

t_mp3stream *
shoutcast_open(const char *url, t_client * clt)
{
  t_mp3stream *pmst;
  pthread_attr_t attr;
  int rc;

  pmst = xmalloc(sizeof(t_mp3stream));
  pmst->url = xstrdup(url);
  pmst->clt = clt;
  pmst->http_buffer_size = 128;
  pmst->http_prebuffer = 25;
  pmst->buffer_length = pmst->http_buffer_size * 1024;
  pmst->prebuffer_length = (pmst->buffer_length * pmst->http_prebuffer) / 100;
  pmst->prebuffering = 1;
  pmst->going = 1;
  pmst->buffer = xmalloc(pmst->buffer_length);

  do_log2(4, pmst->clt->fd,
	  "creating thread to serve Shoutcast stream '%s'\n", pmst->url);
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  rc = pthread_create(&pmst->thread, &attr, http_buffer_loop, pmst);
  if (rc)
    {
      do_log(1, "%s: pthread_create() failed: %s\n", __func__, strerror(rc));
      pmst->thread = (pthread_t) 0;
      shoutcast_close(pmst);
      pmst = NULL;
    }
  return pmst;
}

void
shoutcast_close(t_mp3stream * pmst)
{
  if (pmst == NULL)
    return;

  if (pmst->thread != (pthread_t) 0)
    {
      int rc;

      do_log2(3, pmst->clt->fd,
	      "%s: canceling thread to serve Shoutcast stream\n", __func__);

      pmst->going = 0;
      rc = pthread_join(pmst->thread, NULL);
      if (rc && rc != EDEADLK)
	{
	  do_log(1, "%s: pthread_join() failed: %s\n", __func__,
		 strerror(rc));
	}
    }

  xfree(pmst->url);
  xfree(pmst->buffer);
  xfree(pmst->icy_name);
  xfree(pmst->icy_title);
  xfree(pmst);
}

int
shoutcast_read(void *data, int length, t_mp3stream * pmst)
{
  int len, cnt, off = 0;

  http_wait_for_data(length, pmst);

  if (!pmst->going)
    return 0;

  len = min(http_used(pmst), length);

  /* http://www.smackfu.com/stuff/programming/shoutcast.html

     Read the data stream as you normally would, keeping a byte count as you
     go. When the number of bytes equals the metadata interval, you will get
     a metadata block. The first part of the block is a length specifier,
     which is the next byte in the stream. This byte will equal the metadata
     length / 16. Multiply by 16 to get the actual metadata length. (Max byte
     size = 255 so metadata max length = 4080.) Now read that many bytes and
     you will have a string containing the metadata. Restart your byte count,
     and repeat. Yay!

     Note that the metadata length is set to 0 most of the time, meaning there
     is no metadata. The metadata is normally sent at two particular places:
     immediately after connecting and when the song changes. It seems as if
     some servers aren't real diligent about the former, so it might take a
     while to get your first block.

     Also, be sure you don't count the bytes in the metadata length field or
     the metadata when you're figuring out whether you've hit the interval.
     Only the MP3 data counts. So you can't be lazy and just keep a total byte
     count and mod it. */

  while (len && http_used(pmst))
    {
      if (pmst->icy_metaint > 0)
	{
	  if ((pmst->buffer_read % pmst->icy_metaint) == 0 &&
	      (pmst->buffer_read > 0))
	    {
	      int meta_len =
		*((unsigned char *)pmst->buffer + pmst->rd_index) * 16;
	      pmst->rd_index = (pmst->rd_index + 1) % pmst->buffer_length;
	      if (meta_len > 0)
		{
		  http_wait_for_data(meta_len, pmst);
		  if (http_used(pmst) >= meta_len)
		    {
		      char *tok, *tokptr;
		      char *meta_data = xmalloc(meta_len);
		      int meta_off = 0;

		      while (meta_len)
			{
			  cnt =
			    min(meta_len,
				pmst->buffer_length - pmst->rd_index);
			  memcpy(meta_data + meta_off,
				 pmst->buffer + pmst->rd_index, cnt);
			  pmst->rd_index =
			    (pmst->rd_index + cnt) % pmst->buffer_length;
			  meta_len -= cnt;
			  meta_off += cnt;
			}
		      for (tok = strtok_r(meta_data, ";", &tokptr);
			   tok != NULL; tok = strtok_r(NULL, ";", &tokptr))
			{
			  do_log2(3, pmst->clt->fd, "%s\n", tok);
			  if (strlen(tok) > 12 &&
			      strcasecmp("StreamTitle=", tok) == 0)
			    {
			      xfree(pmst->icy_title);
			      pmst->icy_title = xstrdup(tok + 12);
			    }
			}
		      xfree(meta_data);
		    }
		}
	      if (!http_used(pmst))
		http_wait_for_data(length - off, pmst);
	    }
	  cnt =
	    min4(len, pmst->buffer_length - pmst->rd_index, http_used(pmst),
		 pmst->icy_metaint - (pmst->buffer_read % pmst->icy_metaint));
	}
      else
	cnt = min3(len, pmst->buffer_length - pmst->rd_index,
		   http_used(pmst));

      memcpy((char *)data + off, pmst->buffer + pmst->rd_index, cnt);
      pmst->rd_index = (pmst->rd_index + cnt) % pmst->buffer_length;
      pmst->buffer_read += cnt;
      len -= cnt;
      off += cnt;
    }

  return off;
}

/* EOF */
